##################################################
# 画像変換制御クラス
#
# Shin.Onda 2023/08//26 all rights reserved.
#
##################################################
import cv2
import numpy as np
import os
import re
import sys
import datetime as dt
import tkinter as tk
import tkinter.filedialog as filedialog
import tkinter.messagebox as messagebox
import tkinter.simpledialog as simpledialog
import glob
import imagezoom
import imageshift
import imagerotate
import imageskew
from imageeffect import ImageEffect

WINDOW_NAME = 'imagetransform'

# 画像のデフォルト
IMAGE_DIR = './'

#読み込み画像ファイルタイプ
#(環境に依存します。必要に応じて変更してください)
READ_EXT = r'/*(\.jpg|\.JPG|\.jpeg|\.JPEG|\.png|\.PNG|\.bmp|\.BMP)$'

class ImageTransformManager :
    # 画像シフトクラスのインスタンス
    imageShift = imageshift.ImageShift()
    # 回転クラスのインスタンス
    imageRotate = imagerotate.ImageRotate()
    # 拡大縮小クラスのインスタンス
    imageZoom = imagezoom.ImageZoom()
    # スキュークラスのインスタンス
    imageSkew = imageskew.ImageSkew()

    #######################
    # メンバ変数
    #######################
    # デフォルトのフォルダ
    defaultdir = IMAGE_DIR
    
    # 画像形式
    paramcodec = 'jpg'

    # 背景色
    paramBack = 'R'

    # 効果
    paramEffect = ''
    effectValue = 10

    # 回転オプション
    rotateOption = False
    # 回転開始角度（°）
    rotateMin = 60
    # 回転終了角度（°）
    rotateMax = 60
    # 回転ステップ
    rotateStep = 10
    
    # スキューオプション
    SkewOption = False
    # スキュー開始角度（°）
    SkewMin = 60
    # スキュー終了角度（°）
    SkewMax = 60
    # スキューステップ
    SkewStep = 10

    # 水平移動オプション
    horShiftOption = False
    # 水平移動サイズ（％最小）
    horShiftMin = 50
    # 水平移動サイズ（％最大）
    horShiftMax = 50
    # 水平移動ステップ
    horShiftStep = 10

    # 垂直移動オプション
    verShiftOption = False
    # 垂直移動サイズ（％最小）
    verShiftMin = 50
    # 垂直移動サイズ（％最大）
    verShiftMax = 50
    # 垂直移動ステップ
    verShiftStep = 10

    # 縮小拡大オプション
    zoomOption = False
    # 縮小拡大サイズ（％最小）
    zoomMin = 50
    # 縮小拡大サイズ（％最大）
    zoomMax = 150
    # 縮小拡大ステップ
    zoomStep = 10

    # 左右引き伸ばしオプション
    horZoomOption = False
    # 縮小拡大サイズ（％最小）
    horZoomMin = 50
    # 縮小拡大サイズ（％最大）
    horZoomMax = 150
    # 縮小拡大ステップ
    horZoomStep = 10

    # 上下引き伸ばしオプション
    verZoomOption = False
    # 縮小拡大サイズ（％最小）
    verZoomMin = 50
    # 縮小拡大サイズ（％最大）
    verZoomMax = 150
    # 縮小拡大ステップ
    verZoomStep = 10

    # アクセサ
    # デフォルトのフォルダ
    def getDefaultDir(self) :
        return self.defaultdir
    def setDefaultDir(self, defaultDir) :
        self.defaultDir = defaultDir
    # 画像形式
    def getParamCodec(self) :
        return self.paramcodec
    def setParamCodec(self, paramcodec) :
        self.paramcodec = paramcodec
    # 背景色
    def getParamBack(self) :
        return self.paramBack
    def setParamBack(self, paramBack) :
        self.paramBack = paramBack
    # 効果
    def getParamEffect(self) :
        return self.paramEffect
    def setParamEffect(self, paramEffect) :
        self.paramEffect = paramEffect
    def getEffectValue(self) :
        return self.effectValue
    def setEffectValue(self, effectValue) :
        self.effectValue = effectValue
    # 回転
    def getRotateOption(self) :
        return self.rotateOption
    def setRotateOption(self, rotateOption) :
        self.rotateOption = rotateOption
    def getRotateMin(self) :
        return self.rotateMin
    def setRotateMin(self, rotateMin) :
        self.rotateMin = rotateMin
    def getRotateMax(self) :
        return self.rotateMax
    def setRotateMax(self, rotateMax) :
        self.rotateMax = rotateMax
    def getRotateStep(self) :
        return self.rotateStep
    def setRotateStep(self, rotateStep) :
        self.rotateStep = rotateStep
    # スキュー
    def getSkewOption(self) :
        return self.SkewOption
    def setSkewOption(self, SkewOption) :
        self.SkewOption = SkewOption
    def getSkewMin(self) :
        return self.SkewMin
    def setSkewMin(self, SkewMin) :
        self.SkewMin = SkewMin
    def getSkewMax(self) :
        return self.SkewMax
    def setSkewMax(self, SkewMax) :
        self.SkewMax = SkewMax
    def getSkewStep(self) :
        return self.SkewStep
    def setSkewStep(self, SkewStep) :
        self.SkewStep = SkewStep
    # 水平移動
    def getHorShiftOption(self) :
        return self.horShiftOption
    def setHorShiftOption(self, horShiftOption) :
        self.horShiftOption = horShiftOption
    def getHorShiftMin(self) :
        return self.horShiftMin
    def setHorShiftMin(self, horShiftMin) :
        self.horShiftMin = horShiftMin
    def getHorShiftMax(self) :
        return self.horShiftMax
    def setHorShiftMax(self, horShiftMax) :
        self.horShiftMax = horShiftMax
    def getHorShiftStep(self) :
        return self.horShiftStep
    def setHorShiftStep(self, horShiftStep) :
        self.horShiftStep = horShiftStep
    # 垂直移動
    def getVerShiftOption(self) :
        return self.verShiftOption
    def setVerShiftOption(self, verShiftOption) :
        self.verShiftOption = verShiftOption
    def getVerShiftMin(self) :
        return self.verShiftMin
    def setVerShiftMin(self, verShiftMin) :
        self.verShiftMin = verShiftMin
    def getVerShiftMax(self) :
        return self.verShiftMax
    def setVerShiftMax(self, verShiftMax) :
        self.verShiftMax = verShiftMax
    def getVerShiftStep(self) :
        return self.verShiftStep
    def setVerShiftStep(self, verShiftStep) :
        self.verShiftStep = verShiftStep
    # 縮小拡大
    def getZoomOption(self) :
        return self.zoomOption
    def setZoomOption(self, zoomOption) :
        self.zoomOption = zoomOption
    def getZoomMin(self) :
        return self.zoomMin
    def setZoomMin(self, zoomMin) :
        self.zoomMin = zoomMin
    def getZoomMax(self) :
        return self.zoomMax
    def setZoomMax(self, zoomMax) :
        self.zoomMax = zoomMax
    def getZoomStep(self) :
        return self.zoomStep
    def setZoomStep(self, zoomStep) :
        self.zoomStep = zoomStep
    # 左右引き伸ばし
    def getHorZoomOption(self) :
        return self.horZoomOption
    def setHorZoomOption(self, horZoomOption) :
        self.horZoomOption = horZoomOption
    def getHorZoomMin(self) :
        return self.horZoomMin
    def setHorZoomMin(self, horZoomMin) :
        self.horZoomMin = horZoomMin
    def getHorZoomMax(self) :
        return self.horZoomMax
    def setHorZoomMax(self, horZoomMax) :
        self.horZoomMax = horZoomMax
    def getHorZoomStep(self) :
        return self.horZoomStep
    def setHorZoomStep(self, horZoomStep) :
        self.horZoomStep = horZoomStep
    # 上下引き伸ばし
    def getVerZoomOption(self) :
        return self.verZoomOption
    def setVerZoomOption(self, verZoomOption) :
        self.verZoomOption = verZoomOption
    def getVerZoomMin(self) :
        return self.verZoomMin
    def setVerZoomMin(self, verZoomMin) :
        self.verZoomMin = verZoomMin
    def getVerZoomMax(self) :
        return self.verZoomMax
    def setVerZoomMax(self, verZoomMax) :
        self.verZoomMax = verZoomMax
    def getVerZoomStep(self) :
        return self.verZoomStep
    def setVerZoomStep(self, verZoomStep) :
        self.verZoomStep = verZoomStep

    #######################
    # ファイル操作
    #######################
    #######################
    # 画像書き込み
    # cv2.imwrite で日本語パスが使用できない問題の対策
    #   filepath    = 書き込み先ファイル名パス名
    #   img         = 書き込むイメージ（フレーム）
    # 戻り値 True 正常、False 異常
    #######################
    def imwrite(self, filepath, img) :
        try:
            # 拡張子
            ext = os.path.splitext(filepath)[1]
            # エンコード
            ret, n = cv2.imencode(ext, img)
            if ret:
                # 画像をファイルに書き込む
                with open(filepath, mode='w+b') as f:
                    n.tofile(f)
            else:
                tk.Tk().withdraw()
                ret = messagebox.showinfo('異常', 'ファイル書き込みに失敗しました')
        except :
            tk.Tk().withdraw()
            ret = messagebox.showinfo('異常', 'ファイル書き込みに失敗しました')
            return False
        return True

    #######################
    # 変換後の画像の書き込み
    #   imageList = イメージのリスト
    #   outputDir = 書き込み先フォルダ
    #   mode      = 変換処理の種類
    #   ext       = 拡張子
    # 戻り値 True 正常、False 異常(中止)
    #######################
    def writeImage(self, imageList, outputDir, orgname, mode, ext) :
        pathformat = '{0}_{1}_{2}{3}'
        ret = True

        for i, newImage in enumerate(imageList) :

            if self.getParamEffect() == 'B' :
                # ぼかし
                newImage = ImageEffect.blur(newImage, self.getEffectValue())
            if self.getParamEffect() == 'G' :
                # グレースケール
                newImage = cv2.cvtColor(newImage, cv2.COLOR_BGR2GRAY)
            if self.getParamEffect() == 'N' :
                # ノイズ
                newImage = ImageEffect.noise(newImage, self.getEffectValue())

            # ウィンドウの存在チェック
            ret = cv2.getWindowProperty(WINDOW_NAME, cv2.WND_PROP_VISIBLE)
            if not ret :
                tk.Tk().withdraw()
                retYes = messagebox.askyesno('確認', '書き込み処理を中止しますか？')
                if retYes :
                    ret = False
                    break
            # 表示
            cv2.imshow(WINDOW_NAME, newImage)
            # 中断チェック
            key = cv2.waitKey(10)
            if key == 27 : # ESCキー
                tk.Tk().withdraw()
                retYes = messagebox.askyesno('確認', '書き込み処理を中止しますか？')
                if retYes :
                    ret = False
                    break
            # 書き込み
            writefilename = pathformat.format(orgname, mode, i, ext)
            outputPath = os.path.join(outputDir, writefilename)
            ret = self.imwrite(outputPath, newImage)
            print(outputPath)

        return ret

    #######################
    # 画像変換と書き込み
    #   orgImage  = 元のイメージ
    #   filepath  = 書き込み先ファイル名パス名
    #   outputDir = 書き込み先フォルダ
    #   effect    = 効果オプション
    # 戻り値 True 正常、False 異常(中止)
    #######################
    def transformAndWriteImage(self, orgImage, filepath, outputDir) :

        # ファイル名フォーマット：元のファイル名にオプションと番号を付加する
        orgname, ext = os.path.splitext(os.path.basename(filepath))
        if len(self.getParamCodec()) > 0 :
            ext = '.' + self.getParamCodec()

        print('Please wait...')

        # 元画像
        imageHorShiftList = [orgImage]
        ret = self.writeImage(imageHorShiftList, outputDir, orgname, 'o', ext) 
        if ret == False :
            return False

        # 回転
        imageRotateList = []
        if self.getRotateOption() :
            imageRotateList = self.imageRotate.rotate(orgImage, self.getParamBack(), self.getRotateMin(), self.getRotateMax(), self.getRotateStep()) 
            ret = self.writeImage(imageRotateList, outputDir, orgname, 'r', ext) 
            if ret == False :
                return False

        # スキュー
        imageSkewList = []
        if self.getSkewOption() :
            imageSkewList = self.imageSkew.horSkew(orgImage, self.getParamBack(), self.getSkewMin(), self.getSkewMax(), self.getSkewStep()) 
            ret = self.writeImage(imageSkewList, outputDir, orgname, 's', ext) 
            if ret == False :
                return False

        # 水平シフト
        imageHorShiftList = []
        if self.getHorShiftOption() :
            imageHorShiftList = self.imageShift.horShift(orgImage, self.getParamBack(), self.getHorShiftMin(), self.getHorShiftMax(), self.getHorShiftStep()) 
            ret = self.writeImage(imageHorShiftList, outputDir, orgname, 'sh', ext) 
            if ret == False :
                return False

        # 垂直シフト
        imageVerShiftList = []
        if self.getVerShiftOption() :
            imageVerShiftList = self.imageShift.verShift(orgImage, self.getParamBack(), self.getVerShiftMin(), self.getVerShiftMax(), self.getVerShiftStep()) 
            ret = self.writeImage(imageVerShiftList, outputDir, orgname, 'sv', ext) 
            if ret == False :
                return False

        # 拡大縮小
        imageZoomList = []
        if self.getZoomOption() :
            imageZoomList = self.imageZoom.zoom(orgImage, self.getParamBack(), self.getZoomMin(), self.getZoomMax(), self.getZoomStep(), 0) 
            ret = self.writeImage(imageZoomList, outputDir, orgname, 'z', ext) 
            if ret == False :
                return False

        # 左右引き伸ばし
        imageHorZoomList = []
        if self.getHorZoomOption() :
            imageHorZoomList = self.imageZoom.zoom(orgImage, self.getParamBack(), self.getHorZoomMin(), self.getHorZoomMax(), self.getHorZoomStep(), 1) 
            ret = self.writeImage(imageHorZoomList, outputDir, orgname, 'zh', ext) 
            if ret == False :
                return False

        # 上下引き伸ばし
        imageVerZoomList = []
        if self.getVerZoomOption() :
            imageVerZoomList = self.imageZoom.zoom(orgImage, self.getParamBack(), self.getVerZoomMin(), self.getVerZoomMax(), self.getVerZoomStep(), 2) 
            ret = self.writeImage(imageVerZoomList, outputDir, orgname, 'zv', ext) 
            if ret == False :
                return False

        return True

    #######################
    # 画像読み込み
    # cv2.imred で日本語パスが使用できない問題の対策
    #   filepath  = 書き込み先ファイル名パス名
    # 戻り値 イメージ
    #######################
    def imread(self, filepath):
        try:
            buf = np.fromfile(filepath, np.uint8)
            if buf is None or len(buf) == 0:
                return None
            img = cv2.imdecode(buf, cv2.IMREAD_UNCHANGED)
            depth = img.ndim
            if depth < 3 :
                # グレースケールの場合
                img = cv2.cvtColor(img, cv2.COLOR_GRAY2BGR)
            else :
                #  32ビットの場合、深度を減らす
                org_h, org_w, clr = img.shape
                if clr >= 4 :
                    img = img[:,:,0:3]
        except :
            return None
        return img

    #######################
    # 画像変換処理
    #######################
    def transformate(self) :
        # 元画像のフォルダを選択するループ
        while True :
            # 元画像のフォルダ選択
            imagedir = filedialog.askdirectory(title="元画像のフォルダを選択", initialdir=self.getDefaultDir())
            print(imagedir)
            self.setDefaultDir(imagedir)
            if len(imagedir) == 0 :
                tk.Tk().withdraw()
                ret = messagebox.askyesno('確認', 'フォルダが選択されていません。終了しますか？')
                if ret :
                    print('Canceled.')
                    return
                else :
                    continue  # フォルダ再選択

            # 対象ファイル選択
            matchpath = glob.glob(os.path.join(imagedir, '*.*'))
            imagefiles = [p for p in matchpath if re.search(READ_EXT, str(p))]
            if len(imagefiles) == 0 :
                tk.Tk().withdraw()
                ret = messagebox.askyesno('確認', '処理できる画像ファイルがありません。終了しますか？')
                if ret :
                    print('Canceled.')
                    return
                else :
                    continue  # フォルダ再選択
            # ループ終了
            break

        # 出力先フォルダを選択するループ
        while True :
            # 出力先のフォルダ選択
            outputDir = filedialog.askdirectory(title="出力先のフォルダを選択", initialdir=imagedir)
            print(outputDir)
            if len(outputDir) == 0 :
                tk.Tk().withdraw()
                ret = messagebox.askyesno('確認', 'フォルダが選択されていません。終了しますか？')
                if ret :
                    print('Canceled.')
                    return
                else :
                    continue  # フォルダ再選択

            # サブフォルダ作成
            time = dt.datetime.now().strftime('%Y%m%d%H%M%S')
            outputDir = os.path.join(outputDir, '{0}'.format(time))
            tk.Tk().withdraw()
            if os.path.isdir(outputDir) :
                msg = 'フォルダ[{0}]に書き込みます。既存ファイルがあれば上書きされます。よろしいですか？'.format(outputDir)
                ret = messagebox.askyesno('確認', msg)
                if not ret :
                    continue  # フォルダ再選択
            else :
                msg = 'フォルダ[{0}]を作成します。よろしいですか？'.format(outputDir)
                ret = messagebox.askyesno('確認', msg)
                if not ret :
                    continue  # フォルダ再選択
                os.makedirs(outputDir)

            # ループ終了
            break

        # 最初の画像を読み込み、パラメータ獲得
        print('Reading...' + imagefiles[0])
        img   = self.imread(imagefiles[0])
        org_h, org_w, clr = img.shape
        print(img.shape)
        print('Reading OK.')

        # 元の画像を表示
        cv2.namedWindow(WINDOW_NAME)

        # 読み込みループ
        img_bk = None
        for imageFilePath in imagefiles :

            # 画像を読み込み
            img   = self.imread(imageFilePath)
            if img is None :
                break  # 次がなければ終了

            # ウィンドウの存在チェック
            ret = cv2.getWindowProperty(WINDOW_NAME, cv2.WND_PROP_VISIBLE)
            if not ret :
                tk.Tk().withdraw()
                retYes = messagebox.askyesno('確認', '書き込み処理を中止しますか？')
                if retYes :
                    ret = False
                    break

            # 元の画像を表示
            cv2.imshow(WINDOW_NAME, img)
            
            # 中断チェック
            key = cv2.waitKey(10)
            if key == 27 : # ESCキー
                tk.Tk().withdraw()
                ret = messagebox.askyesno('確認', '書き込み処理を中止しますか？')
                if ret :
                    break

            # 画像処理と書き込み
            ret = self.transformAndWriteImage(img, imageFilePath, outputDir)
            if not ret :
                break
                
            ## -- ループ 終了 --

        ## 終了処理
        cv2.destroyAllWindows()

